﻿/* Copyright 2008-10 Jarrett Vance
 * see http://www.codeplex.com/blogsvc/license */
namespace AtomSite.WebCore
{
  using System;
  using System.Linq;
  using System.Net;
  using System.Security;
  using System.Web.Mvc;
  using System.Web.Routing;
  using AtomSite.Domain;
  using AtomSite.Repository;
  using StructureMap;
  using System.Collections.Generic;

  /// <summary>
  /// This is the base controller to be used by all other controllers.  It provides
  /// a way to populate the model with common services and objects.
  /// </summary>
  public abstract class BaseController : Controller
  {
    protected const int SEC = 1;
    protected const int MIN = 60 * SEC;

    private AppService appService = null;
    public AppService AppService { get { return appService; } }

    public ILogService LogService { get; protected set; }
    public IAuthorizeService AuthorizeService { get; protected set; }

    public AppWorkspace Workspace { get { return !Scope.IsEntireSite ? appService.GetWorkspace(Scope.Workspace) : null; } }
    public AppCollection Collection { get { return Scope.IsCollection ? appService.GetCollection(Scope.Workspace, Scope.Collection) : null; } }
    public Id EntryId { get; protected set; }
    public Scope Scope { get; protected set; }

    /// <summary>
    /// Determines dynamically the collection, workspace, entry, and theme
    /// for the current request.
    /// </summary>
    /// <param name="ctx"></param>
    protected override void Execute(RequestContext ctx)
    {
      IContainer container = (IContainer)ctx.HttpContext.Application["Container"];
      appService = container.GetInstance<IAppServiceRepository>().GetService();
      LogService = container.GetInstance<ILogService>();
      AuthorizeService = container.GetInstance<IAuthorizeService>();
      //determine resource context based on routing data
      IRouteService r = RouteService.GetCurrent(ctx);
      if (EntryId == null) EntryId = r.GetEntryId();
      Scope = r.GetScope();
      base.Execute(ctx);

    }

    protected virtual ViewResult View(BaseModel model)
    {
      return View(null, model);
    }

    protected virtual ViewResult View(string viewName, string masterViewName, BaseModel model)
    {
      SetModel(model);
      return base.View(viewName, masterViewName, model);
    }

    protected virtual ViewResult View(string viewName, BaseModel model)
    {
      SetModel(model);
      if (viewName != null)
        return base.View(viewName, model);
      else
        return base.View(model);
    }

    protected virtual PartialViewResult PartialView(BaseModel model)
    {
      return PartialView(null, model);
    }

    protected virtual PartialViewResult PartialView(string viewName, BaseModel model)
    {
      SetModel(model);
      if (viewName != null)
        return base.PartialView(viewName, model);
      else
        return base.PartialView(model);
    }

    public virtual void SetModel(BaseModel model)
    {
      model.Logger = LogService;
      model.User = User.Identity as User;
      model.AuthorizeService = AuthorizeService;
      if (model.Service == null) model.Service = AppService;
      if (model.Workspace == null) model.Workspace = Workspace;
      if (model.Collection == null) model.Collection = Collection;
      if (model.EntryId == null) model.EntryId = EntryId;
      model.Scope = Scope;
    }

    /// <summary>
    /// Gets the slug set in the header.
    /// </summary>
    /// <returns></returns>
    protected string GetSlug()
    {
      return HttpContext.Request.Headers["Slug"];
    }

    /// <summary>
    /// Determines if the etag does not match in the header
    /// </summary>
    /// <param name="etag"></param>
    /// <returns></returns>
    protected bool NoneMatch(string etag)
    {
      string match = HttpContext.Request.Headers["If-None-Match"];
      return match != etag;
    }

    /// <summary>
    /// Determines if the etag matches in the header
    /// </summary>
    /// <param name="etag"></param>
    /// <returns></returns>
    protected bool Match(string etag)
    {
      string match = HttpContext.Request.Headers["If-Match"];
      return !(!string.IsNullOrEmpty(match) && match != "*" && match != etag);
    }
    
    /// <summary>
    /// Gets a feed of the given type (rss or atom) and remove's emails
    /// to prevent spam
    /// </summary>
    /// <param name="feed"></param>
    /// <param name="type"></param>
    /// <returns></returns>
    protected ActionResult GetFeedResult(AtomFeed feed, string type)
    {
      //remove email's to prevent spam
      foreach (AtomPerson person in feed.People.Concat(feed.Entries.SelectMany(e => e.People)))
      {
        person.Email = null; //string.Empty;
      }

      if (type == "atom")
      {
        return new XmlWriterResult((w) => feed.Xml.WriteTo(w))
        {
          ContentType = Atom.ContentTypeFeed
        };
      }
      else if (type == "rss")
      {
        return new XmlWriterResult((w) => feed.WriteRssTo(w))
        {
          ContentType = "text/xml"
        };
      }
      else
      {
        throw new ResourceNotFoundException(type + " feed", "requested");
      }
    }


    /// <summary>
    /// Returns an AtomPub compliant error based on the exception.
    /// </summary>
    /// <param name="ex"></param>
    /// <returns></returns>
    protected ActionResult GetErrorResult(Exception ex)
    {
      LogService.Error(ex);
      if (ex is ResourceNotFoundException)
      {
        return new HttpResult()
        {
          StatusCode = HttpStatusCode.NotFound,
          StatusDescription = ex.Message
        };
      }
      else if (ex is UserNotAuthenticatedException || ex is UserNotAuthorizedException)
      {
        HttpResult result = new HttpResult()
        {
          StatusCode = HttpStatusCode.Unauthorized,
          StatusDescription = ex.Message
        };
        //let the forms auth bypass module know to remove the redirect to login form
        result.Items["FormsAuthBypass"] = true;
        return result;
      }
      else if (ex is SecurityException)
      {
        return new HttpResult()
        {
          StatusCode = HttpStatusCode.Forbidden,
          StatusDescription = ex.Message
        };
      }
      else if (ex is InvalidContentTypeException)
      {
        return new HttpResult()
        {
          StatusCode = HttpStatusCode.NotAcceptable,
          StatusDescription = ex.Message
        };
      }
      else if (ex is BaseException)
      {
        return new HttpResult()
        {
          StatusCode = HttpStatusCode.BadRequest,
          ContentType = "text/plain",
          Content = ex.Message
        };
      }
      else
      {
        if (HttpContext.IsDebuggingEnabled)
        {
          return new HttpResult()
          {
            StatusCode = HttpStatusCode.InternalServerError,
            ContentType = "text/plain",
            Content = ex.ToString()
          };
        }
        else
        {
          return new HttpResult()
          {
            StatusCode = HttpStatusCode.InternalServerError,
            StatusDescription = "An error occurred. Please contact the " +
              "administrator if the problem persists."
          };
        }
      }
    }

  }
}
