﻿/* Copyright 2008-10 Jarrett Vance
 * see http://www.codeplex.com/blogsvc/license */
namespace AtomSite.WebCore
{
  using System;
  using System.Collections.Generic;
  using System.Linq;
  using System.Web;
  using System.Web.Mvc;
  using AtomSite.Domain;
  using System.Web.Routing;

  public interface IWidget
  {
    string Name { get; }
    string Description { get; }
    IEnumerable<Asset> Assets { get; }
    bool IsEnabled(BaseModel baseModel, Include include);
    void Render(ViewContext ctx, Include include);
    SupportedScopes SupportedScopes { get; }
    void UpdatePageModel(PageModel pageModel, Include include);
  }

  public abstract class BaseWidget : IWidget
  {
    public virtual string Name { get; private set; }
    public virtual string Description { get; set; }
    public virtual IEnumerable<Asset> Assets { get; set; }
    public SupportedScopes SupportedScopes { get; set; }
    public virtual string TailScript { get; set; }
    public BaseWidget(string name)
    {
      this.Name = name;
      this.Assets = new Asset[]{};
      this.TailScript = string.Empty;
    }

    protected virtual Asset AddAsset(string assetName)
    {
      var asset = new Asset(assetName) { AssetScope = AssetScope.Widget };
      Assets = Assets.Concat(new[] {asset});
      return asset;
    }
    protected virtual Asset AddAsset(string assetName, string assetGroup)
    {
      var asset = new Asset(assetName, assetGroup, AssetScope.Widget);
      Assets = Assets.Concat(new[] { asset });
      return asset;
    }

    public virtual bool IsEnabled(BaseModel baseModel, Include include)
    {
      return true;
    }

    public abstract void Render(ViewContext ctx, Include include);

    public virtual void UpdatePageModel(PageModel pageModel, Include include)
    {
      foreach (Asset asset in Assets) pageModel.Assets.Add(asset);
      if (TailScript != null) pageModel.AddToTailScript(TailScript);
    }
  }

  public class ViewWidget : BaseWidget
  {
    public ViewWidget(string name) : base(name) { }

    public override void Render(ViewContext ctx, Include include)
    {
      HtmlHelper helper = new HtmlHelper(ctx, new ViewDataContainer() { ViewData = ctx.ViewData });
      System.Web.Mvc.Html.RenderPartialExtensions.RenderPartial(helper, Name);
    }

    class ViewDataContainer : IViewDataContainer
    {
      public ViewDataDictionary ViewData { get; set; }
    }
  }
  
  public class CompositeWidget : BaseWidget
  {
    protected string Controller { get; private set; }
    protected string Action { get; private set; }

    public CompositeWidget(string name, string controller, string action)
      : base(name)
    {
      this.Controller = controller;
      this.Action = action;
    }

    public override void Render(ViewContext ctx, Include include)
    {
      //TODO: do we need to clone routedata?
      ctx.RouteData.Values["controller"] = Controller;
      ctx.RouteData.Values["action"] = Action;
      ctx.RouteData.Values["include"] = include;
      ctx.RouteData.Values["widgetName"] = this.Name;
      IHttpHandler handler = new MvcHandler(ctx.RequestContext);
      handler.ProcessRequest(System.Web.HttpContext.Current);
    }
  }
}
